// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file ParallelDomainBase.h */

#ifndef AMDIS_PARALLELDOMAINBASE_H
#define AMDIS_PARALLELDOMAINBASE_H


#include <map>
#include <set>
#include <vector>

#include "ProblemTimeInterface.h"
#include "ProblemIterationInterface.h"
#include "FiniteElemSpace.h"
#include "AdaptInfo.h"
#include "InteriorBoundary.h"
#include "Serializer.h"
#include "AMDiS_fwd.h"

#include "petsc.h"
#include "petscsys.h"
#include "petscao.h"
#include "mpi.h"

#include "Global.h"

namespace AMDiS {

  struct DofPtrSortFct {
    bool operator() (const DegreeOfFreedom *dof0, const DegreeOfFreedom *dof1) 
    {
      return (*dof0 < *dof1);
    }
  };
   
  class ParMetisPartitioner;

  class ParallelDomainBase : public ProblemIterationInterface,
			     public ProblemTimeInterface
  {
  private:
    /// Defines a mapping type from DOFs to rank numbers.
    typedef std::map<const DegreeOfFreedom*, int> DofToRank;

    /// Defines a mapping type from DOFs to a set of rank numbers.
    typedef std::map<const DegreeOfFreedom*, std::set<int> > DofToPartitions;

    /// Defines a mapping type from rank numbers to sets of DOFs.
    typedef std::map<int, DofContainer> RankToDofContainer;

    /// Defines a mapping type from DOF indices to DOF indices.
    typedef std::map<DegreeOfFreedom, DegreeOfFreedom> DofMapping;

    /// Defines a mapping type from DOFs to boolean values.
    typedef std::map<const DegreeOfFreedom*, bool> DofToBool;

    /// Defines a mapping type from DOF indices to boolean values.
    typedef std::map<DegreeOfFreedom, bool> DofIndexToBool;

    /// Defines a mapping type from rank numbers to sets of coordinates.
    typedef std::map<int, std::vector<WorldVector<double> > > RankToCoords;

    /// Forward type (it maps rank numbers to the interior boundary objects).
    typedef InteriorBoundary::RankToBoundMap RankToBoundMap;

    typedef std::map<int, DofContainer> ElementIdxToDofs;

    typedef std::map<const DegreeOfFreedom*, DegreeOfFreedom> DofIndexMap;

    typedef std::map<DegreeOfFreedom, std::set<DegreeOfFreedom> > PeriodicDofMap;

  public:
    ParallelDomainBase(ProblemIterationInterface *iterationIF,
		       ProblemTimeInterface *timeIF,
		       FiniteElemSpace *feSpace,
		       RefinementManager *refineManager);

    virtual ~ParallelDomainBase() {}

    virtual void initParallelization(AdaptInfo *adaptInfo);

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    void updateDofAdmins();    

    /** \brief
     * Test, if the mesh consists of macro elements only. The mesh partitioning of
     * the parallelization works for macro meshes only and would fail, if the mesh
     * is already refined in some way. Therefore, this function will exit the program
     * if it finds a non macro element in the mesh.
     */
    void testForMacroMesh();

    /// Set for each element on the partitioning level the number of leaf elements.
    double setElemWeights(AdaptInfo *adaptInfo);

    void partitionMesh(AdaptInfo *adaptInfo);

    virtual void setTime(AdaptInfo *adaptInfo) 
    {
      if (timeIF) 
	timeIF->setTime(adaptInfo);      
    }

    virtual void initTimestep(AdaptInfo *adaptInfo) 
    {
      if (timeIF) 
	timeIF->initTimestep(adaptInfo);
    }

    virtual void closeTimestep(AdaptInfo *adaptInfo) 
    {
      if (timeIF) 
	timeIF->closeTimestep(adaptInfo);
    }

    virtual void solveInitialProblem(AdaptInfo *adaptInfo) 
    {     
      if (timeIF)
	timeIF->solveInitialProblem(adaptInfo);
    }
  
    virtual void transferInitialSolution(AdaptInfo *adaptInfo) 
    {
      if (timeIF) 
	timeIF->transferInitialSolution(adaptInfo);
    }

    virtual void beginIteration(AdaptInfo *adaptInfo) 
    {
      iterationIF->beginIteration(adaptInfo);
    }

    virtual Flag oneIteration(AdaptInfo *adaptInfo, Flag toDo = FULL_ITERATION);

    virtual void endIteration(AdaptInfo *adaptInfo) 
    {
      iterationIF->endIteration(adaptInfo);
    }

    virtual void solve() {}

    virtual int getNumProblems() 
    {
      return 0;
    }

    inline virtual std::string getName() 
    { 
      return name; 
    }

    /// Returns \ref nRankDOFs, the number of DOFs in the rank mesh.
    int getNumberRankDofs() 
    {
      return nRankDofs;
    }

    void fillPetscMatrix(DOFMatrix *mat, DOFVector<double> *vec);

    void fillPetscMatrix(Matrix<DOFMatrix*> *mat, SystemVector *vec);

    void solvePetscMatrix(DOFVector<double> *vec);

    void solvePetscMatrix(SystemVector &vec);

    virtual ProblemStatBase *getProblem(int number = 0) 
    {
      return NULL;
    }

    // Writes all data of this object to an output stream.
    virtual void serialize(std::ostream &out);

    // Reads the object data from an input stream.
    virtual void deserialize(std::istream &in);

  protected:
    /** \brief
     * Determine the interior boundaries, i.e. boundaries between ranks, and store
     * all information about them in \ref interiorBoundary.
     */
    void createInteriorBoundaryInfo();

    /// Removes all macro elements from the mesh that are not part of ranks partition.
    void removeMacroElements();


    /** \brief
     * Creates from a macro mesh a correct local and global DOF index numbering.
     *
     * \param[out] rankDOFs      Returns all DOFs from the macro mesh, which are owned
     *                           by the rank after partitioning the macro mesh.
     * \param[out] nRankDOFs     Number of DOFs owned by rank.
     * \param[out] nOverallDOFs  Number of all DOFs in macro mesh.
     */
    void createLocalGlobalNumbering(DofContainer& rankDOFs,
				    int& nRankDOFs, 
				    int& nOverallDOFs);

    void updateLocalGlobalNumbering(int& nRankDOFs, int& nOverallDOFs);

    /** \brief
     * Creates to all dofs in rank's partition that are on a periodic boundary the
     * mapping from dof index to the other periodic dof indices. This information
     * is stored in \ref periodicDof.
     */
    void createPeriodicMap();

    /** \brief
     * This function create new mappings from local to global indices, 
     * \ref mapLocalGlobalDOFs, and from local to dof indices, \ref mapLocalToDofIndex.
     * Furthermore, using the first argument the dof indices in ranks partition are
     * changed.
     * 
     * \param[in] rankDofsNewLocalIndex       Map from dof pointers of all dofs in rank
     *                                        to new dof indices.
     * \param[in] rankOwnedDofsNewLocalIndex  Map from dof pointers of dofs owned by
     *                                        the rank to the new local index.
     * \param[in] rankDofsNewGlobalIndex      Map from dof pointers of all dofs in rank
     *                                        to the new global index.
     */
    void createLocalMappings(DofIndexMap &rankDofsNewLocalIndex,
			     DofIndexMap &rankOwnedDofsNewLocalIndex,
			     DofIndexMap &rankDofsNewGlobalIndex);

    /** \brief
     * This function traverses the whole mesh, i.e. before it is really partitioned,
     * and collects information about which DOF corresponds to which rank. Can only
     * be used, if \ref partitionVec is set correctly. This is only the case, when
     * the macro mesh is partitioned.
     *
     * \param[out] partionDOFs   Stores to each DOF pointer the set of ranks the DOF is
     *                           part of.
     * \param[out] rankDOFs      Stores all rank DOFs.
     * \param[out] boundaryDOFs  Stores all DOFs in ranks partition that are on an 
     *                           interior boundary but correspond to another rank.
     */
    void createDofMemberInfo(DofToPartitions& partitionDofs,
			     DofContainer& rankOwnedDofs,
			     DofContainer& rankAllDofs,
			     DofToRank& boundaryDofs,
			     DofToBool& vertexDof);

    /** \brief
     * Takes a dof matrix and sends the values to the global petsc matrix.
     */
    void setDofMatrix(DOFMatrix* mat, int dispMult = 1, 
		      int dispAddRow = 0, int dispAddCol = 0);

    /** \brief
     * Takes a dof vector and sends its values to a given petsc vector.
     */
    void setDofVector(Vec& petscVec, DOFVector<double>* vec, 
		      int disMult = 1, int dispAdd = 0);

    void dbgCreateElementMap(ElementIdxToDofs &elMap);
    
    void dbgTestElementMap(ElementIdxToDofs &elMap);

    void dbgTestInteriorBoundary();
     
    /** \brief
     * This function is used for debugging only. It traverses all interior boundaries
     * and compares the dof indices on them with the dof indices of the boundarys
     * neighbours. The function fails, when dof indices on an interior boundary does
     * not fit together.
     *
     * \param  printCoords   If true, the coords of all common dofs are printed to
     *                       the screen.
     */
    void dbgTestCommonDofs(bool printCoords = false);

    /** \brief
     * This function is used for debugging only. It prints all information from
     * the local to global dof mapping, see \ref mapLocalGlobalDOFs.
     *
     * \param  rank  If specified, only the information from the given rank is printed.
     */
    void printMapLocalGlobal(int rank = -1);

    /** \brief
     * This function is used for debugging only. It prints all information about
     * the periodic mapping of dofs, that are on periodic boundaries.
     *
     * \param  rank  If specified, only the information from the given rank is printed.
     */
    void printMapPeriodic(int rank = -1);

    /** \brief
     * This function is used for debugging only. It prints information about dofs
     * in rank's partition.
     *
     * \param  rank         If specified, only the information from the given 
     *                      rank is printed.
     * \param  rankDofs     List of all dofs in ranks partition that are owned by rank.
     * \param  rankAllDofs  List of all dofs in ranks partition.
     */
    void printRankDofs(int rank, DofContainer& rankDofs, DofContainer& rankAllDofs);

    /** \brief
     * This functions create a Paraview file with the macro mesh where the elements
     * are colored by the partition they are part of. This function can be used for
     * debugging.
     */
    void writePartitioningMesh(std::string filename);

    /** \brief
     * This function must be used if the values of a SystemVector must be synchronised
     * over all ranks. That means, that each rank sends the values of the DOFs, which
     * are owned by the rank and lie on an interior bounday, to all other ranks also
     * having this DOF.
     *
     * This function must be used, for example, after the lineary system is solved, or
     * after the SystemVector is set by some user defined functions, e.g., initial
     * solution functions.
     */
    void synchVectors(SystemVector &vec);

    /// Writes a vector of dof pointers to an output stream.
    void serialize(std::ostream &out, DofContainer &data);

    /// Reads a vector of dof pointers from an input stream.
    void deserialize(std::istream &in, DofContainer &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap);

    /// Writes a \ref RankToDofContainer to an output stream.
    void serialize(std::ostream &out, RankToDofContainer &data);

    /// Reads a \ref RankToDofContainer from an input stream.
    void deserialize(std::istream &in, RankToDofContainer &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap);

    /// Writes a periodic dof mapping to an output stream.
    void serialize(std::ostream &out, PeriodicDofMap &data);

    /// Reads a periodic dof mapping from an input stream.
    void deserialize(std::istream &in, PeriodicDofMap &data);

    /// Writes a mapping from dof pointers to some values to an output stream.
    template<typename T>
    void serialize(std::ostream &out, std::map<const DegreeOfFreedom*, T> &data)
    {
      int mapSize = data.size();
      SerUtil::serialize(out, mapSize);
      for (typename std::map<const DegreeOfFreedom*, T>::iterator it = data.begin();
	   it != data.end(); ++it) {
	int v1 = (*(it->first));
	T v2 = it->second;
	SerUtil::serialize(out, v1);
	SerUtil::serialize(out, v2);
      }
    }

    /// Reads a mapping from dof pointer to some values from an input stream.
    template<typename T>
    void deserialize(std::istream &in, std::map<const DegreeOfFreedom*, T> &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap)
    {
      FUNCNAME("ParallelDomainBase::deserialize()");

      int mapSize = 0;
      SerUtil::deserialize(in, mapSize);
      for (int i = 0; i < mapSize; i++) {
	int v1 = 0;
	T v2;
	SerUtil::deserialize(in, v1);
	SerUtil::deserialize(in, v2);

	TEST_EXIT_DBG(dofMap.count(v1) != 0)("Cannot find DOF %d in map!\n", v1);

	data[dofMap[v1]] = v2;
      }
    }
		        
    inline void orderDofs(const DegreeOfFreedom* dof0,
			  const DegreeOfFreedom* dof1,
			  const DegreeOfFreedom* dof2,
			  DofContainer &vec)
    {
      DofPtrSortFct dofPtrSort;
      vec.resize(3);
      vec[0] = dof0; 
      vec[1] = dof1; 
      vec[2] = dof2;
      sort(vec.begin(), vec.end(), dofPtrSort);
    }

    inline void orderDofs(const DegreeOfFreedom* dof0,
			  const DegreeOfFreedom* dof1,
			  const DegreeOfFreedom* dof2,
			  const DegreeOfFreedom* dof3,
			  DofContainer &vec)
    {
      DofPtrSortFct dofPtrSort;
      vec.resize(4);
      vec[0] = dof0; 
      vec[1] = dof1; 
      vec[2] = dof2;
      vec[3] = dof3;
      sort(vec.begin(), vec.end(), dofPtrSort);
    }

    inline void printColValues(int row,
			       std::vector<int>& cols,
			       std::vector<double>& values)
    {
      for (int i = 0; i < static_cast<int>(cols.size()); i++)
	std::cout << "Mat[" << row  << "][" << cols[i] << "] = " << values[i] << "\n";
    }

  protected:
    ///
    ProblemIterationInterface *iterationIF;

    ///
    ProblemTimeInterface *timeIF;

    /// The rank of the current process.
    int mpiRank;

    /// Overall number of processes.
    int mpiSize;

    /** \brief
     * MPI communicator collected all processes, which should
     * be used for calculation. The Debug procces is not included
     * in this communicator.
     */
    MPI::Intracomm mpiComm;

    /// Name of the problem (as used in the init files)
    std::string name;

    /// Finite element space of the problem.
    FiniteElemSpace *feSpace;

    /// Mesh of the problem.
    Mesh *mesh;

    /// Info level.
    int info;

    /// Refinement manager for the mesh.
    RefinementManager *refinementManager;

    /// Pointer to the paritioner which is used to devide a mesh into partitions.
    ParMetisPartitioner *partitioner;

    /// Weights for the elements, i.e., the number of leaf elements within this element.
    std::map<int, double> elemWeights;

    /// Is true, if the mesh was not partitioned before, otherwise it's false.
    bool initialPartitionMesh;

    /** \brief
     * Stores to every coarse element index the number of the partition it 
     * corresponds to.
     */
    std::map<int, int> partitionVec;

    /** \brief
     * Stores an old partitioning of elements. To every element index the number
     * of the parition it corresponds to is stored.
     */
    std::map<int, int> oldPartitionVec;    

    /// Petsc's matrix structure.
    Mat petscMatrix;

    /** \brief
     * Petsc's vector structures for the rhs vector, the solution vector and a
     * temporary vector for calculating the final residuum.
     */
    Vec petscRhsVec, petscSolVec, petscTmpVec;
    
    /// Number of DOFs in the rank mesh.
    int nRankDofs;

    /** \brief 
     * Defines the interior boundaries of the domain that result from partitioning
     * the whole mesh. Contains only the boundaries, which are owned by the rank, i.e.,
     * the object gives for every neighbour rank i the boundaries this rank owns and 
     * shares with rank i.
     */
    InteriorBoundary myIntBoundary;
    
    /** \brief
     * Defines the interior boundaries of the domain that result from partitioning
     * the whole mesh. Contains only the boundaries, which are not owned by the rank,
     * i.e., the object gives for every neighbour rank i the boundaries that are
     * owned by rank i and are shared with this rank.
     */
    InteriorBoundary otherIntBoundary;

    /** \brief
     *
     */
    InteriorBoundary periodicBoundary;

    /** \brief
     * This map contains for each rank the list of dofs the current rank must send
     * to exchange solution dofs at the interior boundaries.
     */
    RankToDofContainer sendDofs;

    /** \brief
     * This map contains for each rank the list of dofs from which the current rank 
     * must receive solution values of dofs at the interior boundaries.
     */
    RankToDofContainer recvDofs;

    /// Maps local to global dof indices.
    DofMapping mapLocalGlobalDOFs;

    /// Maps local dof indices to real dof indices.
    DofMapping mapLocalToDofIndex;  

    /** \brief
     * Maps all DOFs in ranks partition to a bool value. If it is true, the DOF is
     * owned by the rank. Otherwise, its an interior boundary DOF that is owned by
     * another rank.
     */
    DofIndexToBool isRankDof;

    /** \brief
     * Maps every dof pointer in ranks macro mesh to a boolean variable indicating 
     * wheather this dof is a vertex dof (true) or not (false).
     */
    DofToBool vertexDof;

    /** \brief
     * If periodic boundaries are used, this map stores to each dof in rank's 
     * partition, that is on periodic boundaries, the corresponding periodic dofs.
     * The mapping is defined by using global dof indices.
     */
    PeriodicDofMap periodicDof;

    /// Is the index of the first row of the linear system, which is owned by the rank.
    int rstart;

    /** \brief
     * Number of components of the equation. Is used to calculate the exact number
     * of rows in the the overall linear system.
     */
    int nComponents;

    /// Number of rows of the whole linear system that are stored on this rank.
    int nRankRows;

    /// Overall number of the rows in the lineary system.
    int nOverallRows;

    /** \brief
     * If the problem definition has been read from a serialization file, this 
     * variable is true, otherwise it is false. This variable is used to stop the
     * initialization function, if the problem definition has already been read from
     * a serialization file.
     */
    bool deserialized;
  };

}

#endif // AMDIS_PARALLELDOMAINBASE_H
